%
% Initialization routines for the isisscripts
% Executed BEFORE the functions are defined
%

#ifeval __get_reference("_Isisscripts_Verbose")==NULL
variable _Isisscripts_Verbose=1;
#endif

if (Undefined_Type==typeof(_Isisscripts_Verbose)) {
    _Isisscripts_Verbose=1;
}

if (_Isisscripts_Verbose>0) {
    vmessage("loading %s", __FILE__);
}

$1 = __FILE__[[:-4]]+".txt";
ifnot(path_is_absolute($1))  $1 = path_concat(getcwd(), $1);
if(stat_file($1)!=NULL)  set_doc_files([$1, get_doc_files()]);

if (_Isisscripts_Verbose>0) {
    message("\nIf you use results obtained with the ISISscripts in a paper please cite as:");
    message("\"This research has made use of ISIS functions (ISISscripts) provided by ");
    message("ECAP/Remeis observatory and MIT (http://www.sternwarte.uni-erlangen.de/isis/).\"\n");
}


provide("isisscripts");

%
% define local paths
%
#ifeval __get_reference("local_paths")==NULL
variable local_paths = @Struct_Type(String_Type[0]);
#endif
define set_local_paths ()
{
variable default_local_paths = struct {
  localmodels =
  ["/data/system/software/localmodels.xspec12/",  % @ Remeis
   getenv("HOME")+"/share/software/localmodels/",
  ],
  RXTE_data =
  [
   "/eu/X-ray/RXTE/FD/",  % @ Remeis
  ],
  RXTE_ASM_data =
  [
   "/eu/X-ray/RXTE/ASM/",  % @ Remeis
  ],
  RXTE_obscat =
  [
   "/eu/X-ray/RXTE/timelines/obscat/",  % @ Remeis
  ],
  MAXI_lightcurves =
  ["/eu/X-ray/MAXI/lc_1orbit/",  % @ Remeis
  ],
  fermi_catalog =
  ["/eu/X-ray/Fermi/catalogs/2FGL_point_source_cat.fits",  % @ Remeis
  ],
  fermi_catalog_3fgl =
  ["/eu/X-ray/Fermi/catalogs/3FGL_cat.fits",% @ Remeis
  ]
};
return struct_combine(default_local_paths,local_paths);

}

local_paths = set_local_paths();

%
% preload XSPEC local models
%


try {require("xspec");} catch AnyError: {;};

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL

%%%%%%%%%%%%%%%%%%%%%
define use_localmodel()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{use_localmodel}
%\synopsis{load (an) xspec local (table) model[s] from local paths}
%\usage{use_localmodel(String_Type model[, model2, ...]);}
%\qualifiers{
%\qualifier{verbose}{: show the isis command[s] to load the xspec local model}
%\qualifier{atable_name}{: name of the atable model can be specifed
%     (default is the given `model` name); it only works if *one* model is
%     given}
%}
%!%-
{
  if(_NARGS==0) { help(_function_name()); return; }

  variable model, models = ();
  loop(_NARGS-1)
  { model = ();
    models = [models, model];
  }
  variable verbose = qualifier_exists("verbose");

  foreach model ([models])
  { 
    variable i;
    variable path;
    try { 
    foreach path ([local_paths.localmodels]) {
      path += model;
      if(verbose)  vmessage(`%% looking for %s`, path);
      if(length(glob(path))>0)
      { variable e;
	try (e)
	{
	  variable table_file = path+"/"+model+".mod";
	  % check if the model is a table model
	  if(open(table_file,O_EXCL) != NULL)
	  { 
	     model = qualifier("atable_name",model);
	     add_atable_model(table_file, model);
	     if(verbose)  vmessage(`%% isis>  add_atable_model("%s", "%s");`, table_file, model);
	  }
	  else
	  { if(verbose)  vmessage(`%% isis>  load_xspec_local_models("%s");`, path);
	    load_xspec_local_models(path);
	  }

	  if(string_match(model, "warmabs"))
	  { if(verbose)  vmessage(`%% isis>  putenv("LHEA_DATA=%s");`, path);
	    putenv("LHEA_DATA="+path);
	    if(verbose)  vmessage(`%% isis>  putenv("WARMABS_DATA=%s");`, path);
	    putenv("WARMABS_DATA="+path);
	    variable pops = glob(path+"/pops.fits*");
	    _for i (0, length(pops)-1, 1)
	      pops[i] = strreplace(pops[i], path+"/", "");
	    pops = pops[array_sort(pops)];
	    vmessage(`%% isis>  putenv("WARMABS_POP=%s");`, strjoin(pops, "|"));
	    if(getenv("WARMABS_POP")==NULL && length(pops)>0)
	      putenv("WARMABS_POP="+pops[0]);
	  }
	  return;
	}
	catch AnyError:  % seems not to catch Link errors, unfortunately
	  if(verbose)  vmessage("%s: an error was caught: %s\n%s", _function_name(), e.descr, e.message);
      }
    }
    } catch NotImplementedError: { vmessage("warning (%s): local_paths.localmodels not defined",_function_name()); };
    if(verbose)  vmessage("No path to local model '%s' found.", model);
  }
}

%
% use local models
%
use_localmodel("tbnew");

#endif
