%%% functions for model and data binning following certain criteria
%%% 
%%%

%%% optimal binning after Kaastra & Bleeker 2016

%%%
private define binWidthKaastraBleeker (numCounts, numResolver)
%%%
%%+
%\function{binWidthKaastraBleeker}
%\synopsis{Caluculate optimized bin width per data resolution}
%\usage{Double_Type[] binWidthKaastraBleeker (Double_Type[] numCounts,
%                                             Double Type numResolver)}
%\description
%    Kaaster & Bleeker (2016) describe an aproximation to data binning to
%    minimize the number of bins by keeping the resolving power for near
%    resolution limit features.
%    
%    The approximation is based on a Monte Carlo approach to band limited
%    data. The result gives the optimized bin width per FWHM for each data
%    channel. It depends on the number of counts for each channel in the
%    FWHM of the instrument <numCounts> and the number of resolution elemets
%    <numResolver> (roughly sum 1/c_i where c_i is the FWHM in channels for
%    channel i)
%\seealso{}
%%-
{
  variable r = 1 + .2*log(numResolver);
  variable t = log(numCounts*r);
  variable d = Double_Type[length(numCounts)]+1.;
  variable n = where(t>2.119);
  d[n] = (0.08 + 7./t[n] + 1.8/(t[n]^2))/(1+5.9/t[n]);
  return d;
}
%%%

%%%
private define FWHMChannelRange (response)
%%%
{
  variable rLen = length(response);
  variable i;
  variable c1 = Double_Type[rLen];
  variable c2 = Double_Type[rLen];
  _for i (0, rLen-1, 1)
    (c1[i], c2[i]) = hist_fwhm_index(response[i], i; interp);

  return (c1, c2);
}

%%%
private define countsInResolver (c1, c2, response, counts)
%%%
%%+
%\function{countsInResolver}
%\synopsis{Calculate approximated number of counts per FWHM}
%\usage{Double_Type[] countsInResolver (Double_Type[] c1,
%                                       Double_Type[] c2,
%                                       Double_Type[][] response,
%                                       Double_Type[] counts)}
%\description
%
%\seealso{}
%%-
{
  variable size = length(response);
  variable h = Double_Type[size], c = Double_Type[size];
  variable i1, i2;
  variable i,array,len;
  _for i (0, size-1, 1) {
    i1 = int(c1[i]);
    i2 = int(c2[i]);
    array = response[i];
    len = length(array);
    c[i] = sum(counts[[i1:i2]]);
    h[i] = sum(array)/sum(array[[(i1-((i1 == 0)?0:1)):(i2+((i2==len-1)?0:1))]]);
  }
  return c*h;
}

%%%
private define buildOptimalBins (response, counts)
%%%
%%+
%\function{buildOptimalBins}
%\synopsis{Calculate optimal binning (index)}
%\usage{(Int_Type[], Int_Type[]) buildOptimalBins (Double_Type[][] response,
%                                                  Double_Type[] counts)}
%\seealso{}
%%-
{
  variable c1,c2;
  variable resWidth;
  variable numResolver;
  variable cnts;
  variable channelWidth;
  variable i, len, mini, idx;
  variable loIdx, hiIdx;

  (c1,c2) = FWHMChannelRange(response);
  resWidth = c2-c1;
  numResolver = sum(1./resWidth);
  cnts = countsInResolver(c1, c2, response, counts);
  channelWidth = int(binWidthKaastraBleeker(cnts, numResolver)*resWidth);
  channelWidth[where(channelWidth<1)] = 1;
  len = length(counts);
  loIdx = {};
  hiIdx = {};

  i = 0;
  while (i < len) {
    idx = [i:int(_min(i+channelWidth[i]-1, len-1))];
    mini = int(min(idx+channelWidth[idx]));
    list_append(loIdx, i);
    list_append(hiIdx, mini);
    i = mini;
  }
  hiIdx[-1] = len-1; % last index can only be size of array
  return list_to_array(loIdx), list_to_array(hiIdx);
};

%%%
define rebin_dataset_optimal (histIdx)
%%%
%!%+
%\function{rebin_dataset_optimal}
%\usage{rebin_dataset_optimal (Int_Type index)
%\altusage{rebin_dataset_optimal (Int_Type[] index)}}
%\synopsis{Rebin datasets addressed by <index> based on Kaastra & Bleeker 2016}
%\seealso{rebin_data}
%!%-
{
  variable response;
  variable counts;
  variable idx;
  variable idxLo;
  variable idxHi;
  variable i;
  variable flip;
  variable mask;

  foreach idx (histIdx) {
    rebin_data(idx, 0); % reset binning, ugly, yes, but necessary
    response = get_dataset_response(idx);
    counts = get_data(idx).value;

    (idxLo, idxHi) = buildOptimalBins (response, counts); %idxLo and idxHi correspond to energy grid

    mask = Int_Type[length(counts)];
    flip = -1;
    _for i (0, length(idxLo)-1, 1) {
      mask[[idxLo[i]:idxHi[i]]] = flip;
      flip *= -1;
    }

    rebin_data(idx, mask); % The binning mask has again applied on a wavelength grid
  }
}
