
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_dataset_response (id)
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_dataset_response}
%\usage{Response = get_dataset_response(id);}
%\synopsis{Compute response of dataset associated to 'id' to delta peak}
%\qualifiers{
%\qualifier{energy}{: If given, return response on energy grid (default: wavelength)}
%}
%\description
%    This function evaluates a delta peak function set at each nominal
%    grid value. The returnend array contains the response for each peak
%    position. Per default the response is returned as 'seen' by ISIS, that
%    means on a wavelength grid. Use the 'energy' qualifier to revers the
%    response.
%
%\seealso{get_rmf_data_grid, assign_rmf, assign_rsp}
%!%-
{
  % check if dataset exists
  if (all(id != all_data()))
    throw UsageError, sprintf("Dataset %d not defined", id);

  variable rmfId = get_data_info(id).rmfs[0]; % get first rmf id from dataset (see meanGridValues)
  variable grid; % grid from rmf
  variable saveIncludeData = all_data(1); % save current excluded/included data
  variable saveNoticeBins = get_data_info(id).notice_list; % save noticed bins for dataset
  variable saveFitFun = get_fit_fun(); % save current fit fun to recover if necessary
  variable meanGridValues; % mean value from the first rmf grid (this is the one used for the data, see assign_rsp)
  variable i; % running index
  variable fitObject; % the fit object
  variable responseResult; % calculated responses
  variable energy = qualifier_exists("energy"); % reverse returning array so it matches energy grid

  % prepare delta eval
  exclude(all_data()); % exclude everything but the data of interest
  include(id);
  notice(id); % notice all data points
  fit_fun("delta(1)");

  % what is the (unmodified) data grid when more than one rmf is in use?
  % take get_data instead?

  % collect all grid means
  grid = get_rmf_data_grid(rmfId);
  if (NULL == grid) % with define_counts get_rmf_data_grid returns NULL use get_data in any case?
    grid = get_data(id);
  meanGridValues = .5*(grid.bin_hi+grid.bin_lo);

  responseResult = Array_Type[length(meanGridValues)];

  fitObject = open_fit(); % open function
  _for i (0, length(responseResult)-1, 1) {
    () = fitObject.eval_statistic([1,meanGridValues[i]]); % eval function with norm=1 and lambda=mean
    if (energy)
      responseResult[i] = reverse(get_model(id).value);
    else
      responseResult[i] = get_model(id).value;
  }
  fitObject.close();

  % rest to initial settings
  notice_list(id,saveNoticeBins);
  exclude(all_data());
  include(saveIncludeData);
  fit_fun(saveFitFun);

  if (energy)
    return reverse(responseResult);
  else
    return responseResult;
}
