private define filterFouQuantForFrequencyRange(fouquan, f_min, f_max) {
  variable filt = where(f_min < fouquan.freq <= f_max);
    if (qualifier("verbose",0)>1){
      vmessage("  Minimal frequency bin %d at %g Hz", filt[0], fouquan.freq[filt[0]]);
      vmessage("  Maximal frequency bin %d at %g Hz", filt[-1], fouquan.freq[filt[-1]]);
      vmessage("  Number of frequency bins: %d", length(filt));
    }
  return struct_filter(fouquan, filt ; copy);
}

private define verifyLightcurveTime(lc1, lc2){
  %% Test whether the time interval of reference and energy-resolved
  %% lightcurve is the same
  if (length(lc1.time)!=length(lc2.time) or (lc1.time[0]!=lc2.time[0]) or (lc1.time[-1]!=lc2.time[-1])){
    exit("ERROR: The time arrays of the reference and small band lightcurve seem to be different!");
  }
}

private define getLightcurve(lc, dt, dimseg){
  return segment_lc_for_psd(fits_read_table(lc), dt, dimseg ;; __qualifiers);
}

private define calculateSummedReferenceLightcurve(lc_list, index){
  %% As reference lightcurve, take the summed lightcurves of all
  %% energy bands but the one we do the cross-spectrum with (the
  %% lightcurve at index, often called the "channel of interest").
  variable channel_of_interest = fits_read_table(lc_list[index]);

  variable jj;
  variable summed_rate = Double_Type[length(channel_of_interest.rate)];
  _for jj (0, length(lc_list)-1, 1){
    if (jj==index){
      continue;
    } else {
      variable lc = fits_read_table(lc_list[jj]);
      verifyLightcurveTime(channel_of_interest, lc);
      summed_rate += lc.rate;
      % Error is not needed for cross-spectrum calculation
    }
  }
  return struct{time=channel_of_interest.time, rate=summed_rate};
}

private define getReferenceLightcurve(lc_ref, cof_lc, index, dt, dimseg){
  % Default: calculate the reference lightcurve as sum of all
  % energy-resolved lightcurves but the current channel-of-interest.
  % The Poisson error will not be correlated and therefore we do not
  % need to subtract off the channel-of-interest. In this case lc_ref
  % is a list of lightcurves.
  %
  % If we use a broad reference lightcurve, which contains the
  % channel-of-interest, we need to subtract off the small band LC
  % from the reference lightcurve. Uttley+14 call this a
  % channel-of-interest (CI) correction. This is needed to account
  % for the Poisson noise which is correlated for both LCs. See
  % also Uttley+14 Sect 3.2. In this case the lc_ref parameter is an
  % already loaded, segmented lightcurve.
  variable rate;
  if (qualifier_exists("lc_ref")==1){
    rate = lc_ref.rate - cof_lc.rate;
  } else {
    variable summed_lc_ref = calculateSummedReferenceLightcurve(lc_ref, index);
    lc_ref = segment_lc_for_psd(summed_lc_ref, dt, dimseg ;; __qualifiers);
    rate = lc_ref.rate;
  }
  verifyLightcurveTime(lc_ref, cof_lc);
  return struct{time=lc_ref.time, rate=rate};
}

define lag_energy_spectrum(lc_list, dimseg) {
%!%+
%\function{lag_energy_spectrum}
%\synopsis{Calculates the time lag of energy-resolved lightcurves}
%\usage{Struct_Type les = lag_energy_spectrum(String_Type[lcs], dimseg)}
%\description
%    * lc_list: Array of file names of the energy-resolved
%               lightcurves (must have same length and time binning)
%    * dimseg:  Segmentation length in bins, needed to segment the
%               lightcurve and do the Fourier calculation
%    
%    returns: lag-energy-spectrum as a struct with fields lag, errlag
%\qualifiers{
%    lc_ref    - Filename of a (broad) reference lightcurve. If not
%                set, the reference lightcurve will consist of the
%                summed up energy lightcurves except for the current 
%                channel-of-interest.
%    dt        - Time resolution of the lightcurve in seconds
%                [default: TIMEDEL keyword] 
%    f_min     - For frequency-resolved. Default: lowest sampled frequency
%                [1/(dt*dimseg)]. Unit: Hz
%    f_max     - See f_min. Default: Nyquist frequency [1/(2*dt)]. Unit: Hz
%    verbose   - Increase verbosity [default=0]
%    outfile   - String_Type or Fits_File_Type. If set, write output
%                either to open FITS file pointer or create new file
%                (which then contains only the current frequency
%                interval)
%    elo, ehi  - Double_Type: Energy grid, only used for writing outfile
%    
%    (Furthermore, the qualifiers of segment_lc_for_psd can be specified)
%}
%\notes
%    * The default error computation is Nowak et al., 1999, ApJ, 510,
%      874 (Sect. 4), i.e., from coherence without bias factor.
%    * All input lightcurves are accounted for their gaps and segmented
%      by segment_lc_for_psd. 
%    * The cross-spectrum (CPD) and lags are calculated with foucalc.
%    * The PSD normalization is fixed to Miyamoto.
%    * The average time lag in the frequency interval is calculated by
%      the normal mean on the imaginary and real part of the
%      cross-spectrum: 
%      atan2(mean(Imag(CPD)), mean(Real(CPD)))/(PI*(f_min+f_max))
%    * Reference lightcurve: For each energy bin, the function takes the
%      summed lightcurves of all but the current energy bands (always
%      excluding the lightcurve for the cross-spectrum).
%    * The function does *not* use any information about the energy
%      grid. This has to be created by yourself (and must be the same as
%      the lightcurves in lc_list)! The elo, ehi qualifiers are only
%      for the records in the output file
%    
%    If outfile is a string, a new FITS file will be created, but you
%    can also give a file pointer and the function will only create a
%    new extension called BAND_<f_min>_<f_max>Hz. This way you can
%    store frequency-resolved lag-energy spectra in one output file by
%    always passing the same file pointer (see example). If you want
%    to add energy information to the output FITS file, you can pass
%    two Double_Type arrays (same length as lc_list) via the elo, ehi
%    qualifiers. The columns of the created FITS file are:
%  
%    ENERGY_LO          - Energy channel (keV), taken from elo qualifier
%    ENERGY_HI          - Energy channel (keV), taken from ehi qualifier
%    MEAN_PHASE         - Phase (radians), computed as atan2(mean(Im(CPD)), mean(Re(CPD)))
%    LAG                - Time lag (s), computed as phase/(PI*(f_min+f_max))
%    LAG_ERR_UTTLEY+14  - Error computed from coherence with bias
%                         Uttley et al., 2014, A&A Rv. 22, 72 (Eq. 11-12)
%    LAG_ERR_NOWAK+99   - Error computed from coherence without bias (default)
%                         Nowak et al., 1999, ApJ, 510, 874 (Sect. 4)
%    LAG_ERR_PROPERR    - Error computed from Gaussian error propagation
%    LAG_ERR_INGRAM19   - (to be implemented)
%    MEAN_REAL/IMAG_CPD - real/imaginary part of the mean of the
%                         segment-averaged cross-spectrum
%    MEAN_REAL/IMAG_CPD_ERR - standard error on the mean of the
%                         real/imaginary segment-averaged cross-spectrum 
%    N_SEGMENTS         - Number of segments averaged over (foucalc.numavgall)
%    N_FREQUENCIES      - Number of frequencies averaged over (as in
%                         "mean of the CPD")
%    MIN/MAX_FREQUENCY  - Minimal/Maximal frequency of the frequency-filtered
%                         Fourier products - this can be different from
%                         f_min due to the discretization of the
%                         *Discrete* Fourier Transform
%    RMS                - Root Mean Squared variability of the
%                         channel-of-interest lightcurve
%    RMS_REF            - RMS variability of the reference lightcurve
%                         
%    The last three fields should be the same for all lightcurves.
%    
%    If you find any bugs or have questions, please contact ole.koenig@fau.de
%\example
%    %% Example for frequency-resolved lag-energy spectrum
%    (flo, fhi) = log_grid(1, 50, n_freqs);  % (Hz)
%    % Energy grid must be the same as extracted lightcurves
%    (elo, ehi) = log_grid(0.5, 10, n_energies);  % (keV)
%    variable fp = fits_open_file("test.fits", "c");
%    _for ii (0, n_freqs-1, 1) {  % going through frequency range
%      variable les = lag_energy_spectrum(lc_names, dimseg
%                                         ; f_min = flo[ii], f_max=fhi[ii],
%                                         verbose=2, outfile=fp, elo=elo, ehi=ehi);
%      hplot_with_err(elo, ehi, les.lag, les.errlag);   
%    }
%    fits_close_file(fp);
%\seealso{segment_lc_for_psd, foucalc, colacal}
%!%-
  variable dt = qualifier("dt", fits_read_key(lc_list[0], "TIMEDEL"));
  variable f_min = qualifier("f_min", 1/(dt*dimseg));
  variable f_max = qualifier("f_max", 1/(2*dt));  % Nyquist frequency
  variable verbose = qualifier("verbose", 0);
  variable normtype = "Miyamoto";
  
  if (dt==NULL) {
    return "ERROR: Couldn't receive the time resolution from TIMEDEL, need dt qualifier!";
  }
  
  if (verbose>0) vmessage("Frequency range %g-%g Hz", f_min, f_max);

  variable N_energies = length(lc_list);
  variable rmss = Double_Type[N_energies];
  variable rms_refs = Double_Type[N_energies];
  variable lags = Double_Type[N_energies];
  variable errlag_nowak99 = Double_Type[N_energies];
  variable errlag_uttley14 = Double_Type[N_energies];

  variable lc_ref;
  if (qualifier_exists("lc_ref")==1){
    lc_ref = getLightcurve(qualifier("lc_ref"), dt, dimseg ;; __qualifiers);
  } else {
    lc_ref = lc_list;
  }
  
  if (qualifier_exists("outfile")==1){
    variable fp;
    if (typeof(qualifier("outfile")) == String_Type){
      fp = fits_open_file(qualifier("outfile"), "c");
    } else {
      fp = qualifier("outfile");
    }
    variable min_freqs = Double_Type[N_energies];
    variable max_freqs = Double_Type[N_energies];
    variable mean_realcpds = Double_Type[N_energies];
    variable mean_realcpd_errs = Double_Type[N_energies];
    variable mean_imagcpds = Double_Type[N_energies];
    variable mean_imagcpd_errs = Double_Type[N_energies];
    variable mean_phases = Double_Type[N_energies];
    variable errlag_properr = Double_Type[N_energies];
    variable errlag_ingram19 = Double_Type[N_energies];
    variable n_segments = Integer_Type[N_energies];
    variable n_frequencies = Integer_Type[N_energies];
  }
  
  %% Go through the energy-resolved lightcurves (channel-of-interests)
  variable ii;
  _for ii (0,N_energies-1,1){ 
    
    variable channel_of_interest = getLightcurve (lc_list[ii], dt, dimseg ;; __qualifiers);
    variable reference = getReferenceLightcurve (lc_ref, channel_of_interest, ii, dt, dimseg ;; __qualifiers);

    variable rms;
    variable res = foucalc( struct{ time  = channel_of_interest.time,
                                    rate1 = reference.rate,
                                    rate2 = channel_of_interest.rate},
			    dimseg ; normtype=normtype, RMS=&rms);
    (rmss[ii], rms_refs[ii]) = rms[0], rms[1];
    
    variable res_f = filterFouQuantForFrequencyRange(res, f_min, f_max ; verbose=verbose);
    
    %% Coherence and lag calculation on averaged cross-spectrum
    variable cola = colacal(0.5*(f_min+f_max),  % mid frequency
			    mean(res_f.realcpd12)+mean(res_f.imagcpd12)*1i,
			    mean(res_f.noicpd12),  % not needed for lag+error
			    mean(res_f.rawpsd1), mean(res_f.rawpsd2), 
			    mean(res_f.noipsd1), mean(res_f.noipsd2),
			    mean(res_f.sigpsd1), mean(res_f.sigpsd2),  % not needed for lag+err
			    res_f.numavgall[0]*length(res_f.freq));  % K*M with M=number of frequencies averaged over and K=number of segments
    (lags[ii], errlag_nowak99[ii], errlag_uttley14[ii]) = cola.lag, cola.errlag, cola.errlag_uttley14;
    
    if (qualifier_exists("outfile")==1){
      variable m_ReCPD = mean(res_f.realcpd12);
      variable m_ReCPD_err = mean(res_f.errrealcpd12);
      variable m_ImCPD = mean(res_f.imagcpd12);
      variable m_ImCPD_err = mean(res_f.errimagcpd12);

      %% Calculation from Gaussian error propagation
      variable m_phase = atan2(m_ImCPD, m_ReCPD);  % radians
      errlag_properr[ii]=1/(PI*(f_min+f_max))*
	sqrt((m_ReCPD/(m_ImCPD^2+m_ReCPD^2))^2*m_ImCPD_err^2 +
	     (m_ImCPD/(m_ImCPD^2+m_ReCPD^2))^2*m_ReCPD_err^2
	    );
      
      %% Implement error calculation from Ingram 2019, Eq. 19

      min_freqs[ii] = min(res_f.freq);
      max_freqs[ii] = max(res_f.freq);
      mean_realcpds[ii] = m_ReCPD;
      mean_realcpd_errs[ii] = m_ReCPD_err;
      mean_imagcpds[ii] = m_ImCPD;
      mean_imagcpd_errs[ii] = m_ImCPD_err;
      mean_phases[ii] = m_phase;
      n_segments[ii] = res_f.numavgall[0];
      n_frequencies[ii] = length(res_f.freq);
    }
  }

  if (qualifier_exists("outfile")==1){
    variable extname = sprintf("BAND_%g_%gHz", f_min, f_max);
    variable sdata = struct{
      "ENERGY_LO"=qualifier("elo", Double_Type[N_energies]),
      "ENERGY_HI"=qualifier("ehi", Double_Type[N_energies]),
      "LAG"=lags, 
      "LAG_ERR_UTTLEY+14"=errlag_uttley14,
      "LAG_ERR_NOWAK+99"=errlag_nowak99, 
      "LAG_ERR_PROPERR"=errlag_properr,
      "LAG_ERR_INGRAM19"=errlag_ingram19,
      "MEAN_IMAG_CPD"=mean_imagcpds,
      "MEAN_IMAG_CPD_ERR"=mean_imagcpd_errs,
      "MEAN_REAL_CPD"=mean_realcpds,
      "MEAN_REAL_CPD_ERR"=mean_realcpd_errs,
      "MEAN_PHASE"=mean_phases,
      "N_SEGMENTS"=n_segments,
      "N_FREQUENCIES"=n_frequencies,
      "MIN_FREQUENCY"=min_freqs,
      "MAX_FREQUENCY"=max_freqs,
      "RMS"=rmss,
      "RMS_REF"=rms_refs
    };
    variable keys = struct{
      F_MIN=f_min, F_MAX=f_max, DIMSEG=dimseg, TIMEDEL=dt, NORMTYPE=normtype
    };
    variable hist = struct{
      history = [sprintf("Wrote lag-energy spectrum in frequency range %f-%fHz", f_min, f_max)],
      comment = ["Time unit is seconds, energy unit is keV, frequency unit is Hertz", 
		 "Analyzed lightcurves:", lc_list]
    };
    fits_write_binary_table(fp, extname, sdata, keys, hist);
    
    if (typeof(qualifier("outfile")) == String_Type) fits_close_file(fp);
  }
  
  return struct{ lag = lags, errlag = errlag_nowak99 };
}
