%%%%%%%%%%%%%%%%%%%%%%%%
define pfold ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pfold}
%\synopsis{folds a lightcurve on a given period}
%\usage{(Struct_Type pp) = pfold(Double_Type t, r, p);
% or (Struct_Type pp) = pfold(Double_Type t, r, p, e); 
% or (Struct_Type pp) = pfold(Double_Type t, p) ; (event data) }
%\qualifiers{
%\qualifier{nbins}{: number of bins for the pulse profile}
%\qualifier{exact}{: calculate the pulse profile in a more exact way, see description.}
%\qualifier{returnlc} {: returns the LC and the corresponding phases and phase bins for every LC bin. Used mainly for debugging.}
%\qualifier{dt} {:exposure of every lightcurve time bin, should be given to ensure correct results.}
%\qualifier{t0} {: set reference time.}
%\qualifier{pdot} {:first derivative of pulse period p (default=0).}
%\qualifier{pddot} {:second derivative of pulse period p (default=0).}
%\qualifier{gti} {:GTIs for event data, given as struct{start=Double_Type, stop=Double_Type}}
%}
%\description
%   Calculates the pulse profile of a given lightcurve, given as time
%   t and rate r for the period p. The function was adopted from
%   the IDL routine of the same name. The GTI filtering is
%   only implemented for eventdata.
%
%   The data arrays time (t), rate (r), and error (r), can also be
%   given as qualifier (t, r, e, respectively), which allows pfold to
%   be run with "array_map_qualifiers" on an array of periods (p).
%   
%   Compared to the similar function sitar_pfold_rate, pfold.sl
%   calculates the errors pp.error (if given as a forth argument) with
%   correct error propagtion, if e is given. When no errors are given,
%   the variance in each bin is used as in sitar_pfold_rate. If the
%   qualifier "dt" is given, the mean count rate in every phase bin is
%   calculated by weighing the countrates by the exposure time, unlike
%   sitar_pfold_rate.
%   
%   If the "exact" qualifier is given, the function takes the exposure
%   time of every time bin into account in the sense that, that a
%   given time bin may overlap over two phase bins. The corresponding
%   exposure time in every phase bin is reduced accordingly.
%   
%   NOTE: the "exact" qualifiers slows the script down, depending on
%   the length of the lightcurve and the number of bins for up to a
%   factor of >10.
%   
%   If "exact" is not given, the script is ~10% slower than
%   sitar_pfold_rate.
%
%   If p is a function of time described by a taylor expansion
%   (p(t) = p0 + pdot/2*t + pddot/6*t^2 + ...), the first 2 derivatives
%   "pdot" and "pddot" can be set as qualifieres.
%
%   %   
%!%-
{
   
  % tic ;
   % disable chatty if not set
   variable chatty = qualifier("chatty",0) ;


   variable t,r = Double_Type[0] ; 
   variable p,e = Double_Type[0]  ;
   variable noerr = 0 ;
   variable eventdata = 0 ;
  
   switch(_NARGS)
     {case 1: (p) = () ;
	t = qualifier("t", Double_Type[0]) ;
	r = qualifier("r", Double_Type[0]) ;
	e = qualifier("e", Double_Type[0]) ;
     }
     {case 2: (t,p) = () ;
     	r = qualifier("r", Double_Type[0]) ;
	e = qualifier("e", Double_Type[0]) ;
     }
    {case 3: (t,r,p) = ();
    	e = qualifier("e", Double_Type[0]) ;
    }
    {case 4: (t,r,p,e) = (); }
    {help(_function_name()); return; }
  
   ifnot (length(e)) {e=Double_Type[length(r)]; noerr  = 1 ; }

   if (length(r)==0) {
      eventdata = 1 ;
      if ( chatty > 0) {print("using event data"); }
   }

   if (length(t)==0)
       {
	  print("ERROR (pfold): Length of time array is zero, aborting!");
	  return;
       }

   variable nbins = qualifier ("nbins", 20) ;
  
   variable pdot = qualifier("pdot", 0) ;
   variable pddot = qualifier("pddot", 0) ;

  %consistency check that t and r are of the same length
   if (length(t) != length(r) and eventdata==0) {message("ERROR (pfold): time and rate array must be of the same length!"); return ; }

  ifnot(qualifier_exists("dt") or eventdata==1){
    message("WARNING (pfold): no dt qualifier given, creating differential grid! Might lead to unexpected results!") ;
  }
   variable dt = qualifier("dt" , make_hi_grid(t)-t) ;

 ifnot (qualifier_exists("gti") or eventdata==0){
       message("WARNING (pfold): using eventdata without gti qualifier, taking time of first and last event. This is most likely NOT what you want!") ;
 }
  variable gti = qualifier("gti", struct{start=min(t), stop=max(t)}) ;

  variable t0 = qualifier ("t0", t[0]);
   t = t - t0 ;
  
   % ()=printf("setting it up took %fsec\n", toc ; ) ;
   
   % tic ;

   % variable phi_lo = orbitalphase(t,    t0, p, pdot, pddot) ;  %phase of start of time bin
   % variable phi_hi = orbitalphase(t+dt, t0, p, pdot, pddot) ;  %phase of end of time bin
   
   % variable phi_lo = p + ( pdot/2. +  pddot*t/6.)*t ;  %phase of start of time bin
   % phi_lo = ( t mod phi_lo ) / phi_lo ;
   % variable phi_hi = p + ( pdot/2. +  pddot*(t+dt)/6.)*(t+dt) ;  %phase of start of time bin
   % phi_hi = ( t mod phi_hi ) / phi_hi ;
   % variable phi_hi = phi_lo ;
   
   % ()=printf("orbitalphase took %fsec\n", toc ; ) ;
   
   % tic ;
   
   variable gphilo, gphihi ; % phasebin grid
   variable rev_lo, rev_hi ; % variable for reverse indices of phase histogram

   (gphilo, gphihi) = linear_grid(0,1,nbins) ;

   % ()=printf("making grid %fsec.\n", toc ; );    
   % tic ;
   
   % variable h_lo= histogram(phi_lo, gphilo, gphihi , &rev_lo) ;
   % variable h_hi= histogram(phi_hi, gphilo, gphihi , &rev_hi) ;
   % ()=printf("histogram %fsec.\n", toc ; );    

   % tic ;
   variable i ;

   variable profile = Double_Type[nbins] ;
   variable proferr = Double_Type[nbins] ;
   variable npts = Double_Type[nbins] ;
   variable ttot = Double_Type[nbins ] ;
   variable dte = 0 ;
   variable ndx ;
  
   variable g, same , diff;
   
   variable d , tofphi ;
  
   variable rlo, rhi ;

   variable phi_lo, phi_hi, h_lo, h_hi, lc ;

  variable tstart, tstop, p_mid, t_mid ;
   % ()=printf("remaining set-up %fsec\n", toc ; ) ;

  variable phtstart ;
  variable phtstop ;
  variable nstart ;
  variable nstop;

   variable tperbin ;
   variable adddt ;
   
  if (eventdata)
    {
       
       variable allphas = -10.[Integer_Type[length(t)]] ; % in case we want to return the phase of each event
       
      _for g (0, length(gti.start)-1, 1) 
	 {
	    
	    adddt=0 ;
        % ()=printf(" ------------- GTI No %d ------------- \n", g);
	tstart = 1.0*gti.start[g]-t0 ;
	tstop = 1.0*gti.stop[g]-t0 ;
	
	% ()=printf("TSTART = '%.7f, TSTOP = %.7f\n", tstart, tstop) ;

	    % ()=printf("using pdot=%.4e and pddot=%.4e\n", pdot, pddot) ;
	    
	%assume that P does NOT change significantly over the GTI,
	%otherwise this code "would be a real pain" (see the IDL
	%routine pfold.pro of the aitlib).
	t_mid = (gti.start[g]+gti.stop[g])/2. -t0 ;
        p_mid  = 1.0*p+(pdot + pddot*(t_mid)/2.)*(t_mid) ;
	    
	% ()=printf("PMID = %.9f\n", p_mid) ;
	   
        variable nper = (tstop-tstart)/p_mid  ;
	    
	% ()=printf("NPER = %.4f\n", nper) ;
	    
	phtstart = (tstart >= 0 ) ? ((tstart) mod p_mid)/p_mid : (1+(tstart mod p_mid)/p_mid)  ;
        phtstop =  (tstop  >= 0 ) ? ((tstop) mod p_mid)/p_mid  : (1+(tstop mod p_mid)/p_mid  ) ;


        % phtstart = (tstart+t0 >= 0 ) ? ((tstart+t0) mod p_mid)/p_mid : (1+(tstart+t0) mod p_mid)/p_mid  ;
        % phtstop =  (tstop+t0  >= 0 ) ? ((tstop+t0) mod p_mid)/p_mid  : (1+(tstop+t0) mod p_mid)/p_mid   ;

	% ()=printf("PHTSTART = %.3f, PHTSTOP = %.3f\n", phtstart, phtstop) ;
	    
	% print(phtstop-phtstart) ;
	% print(nper);
	% print("______-----___-__-__") ;
	nstart = phtstart*nbins ;
	nstop = phtstop*nbins ;

	% ()=printf("NSTART = %.3f, NSTOP = %.3f\n", nstart,nstop) ;
	    
	tperbin = p_mid/nbins ;
	    
	% ()=printf("tperbin = %.4f\n", tperbin) ;
	% ()=printf("1/nbins = %.4f\n", 1./nbins) ;
	    
	%first take care only of whol phase-bins (partial phase bins
	%can be taken care of in the end)
	 
	%case that tstop and tstart are within one period
	if (nper < 1) 
	  {
	  % vmessage("case nper < 1") ;
	  %case that they are less than a phase-bin separated
	  if (nper < 1./nbins)
	    {
	      % vmessage("case nper < 1/nbins") ;
              % case that tstop and tstart are also within one phasebin
	      if (int(nstop)==int(nstart))
		 {
		    ttot[[int(nstop)]] += (nstop-nstart)*tperbin;
		    % ()=printf("added %.12fs to bin %d\n", (nstop-nstart)*tperbin, int(nstop)) ;
		    adddt+= (nstop-nstart)*tperbin;
		 }
	    }
	  %tstop and tstart are in different phase bins
	  else 
	    {
       	    % vmessage("nper <1, nper > 1/nbins") ;
	    % within one pulse
	    if (nstart<nstop)
	       {
		  % vmessage("both n's in one pulse") ;
	      ttot[[int(ceil(nstart)):int(floor(nstop))-1]] += tperbin ;
		   % ()=printf("added %.12fs to bin %d:%d\n", tperbin, int(ceil(nstart)), int(floor(nstop)-1)) ;
		  adddt+= tperbin*length([int(ceil(nstart)):int(floor(nstop))-1]);
	       }
	     % stretching over the pulse border (but still nper<1)
	     else
		 {
		    
		 % vmessage("n's in different pulses") ;
	      ttot[[[0:int(floor(nstop)-1)],[int(ceil(nstart)):nbins-1]]] += tperbin ;
	       % ()=printf("added %.12fs to bins %d:%d and %d:%d\n", tperbin, 0,int(floor(nstop)-1),
			 % int(ceil(nstart)), nbins-1) ;
		   adddt+=tperbin*length([[0:int(floor(nstop)-1)],[int(ceil(nstart)):nbins-1]]) ;
		 }
             }
	  }
	% tstop and tstart are more than 1 period apart
	else 
	  {
	    % vmessage("case nper >=1") ;
	   %calculate exposure time per bin
	   %check how many complete periods fit into gti interval and
	   %calculate their overall exposure time
  	
	   %again, taking care only care of whole periods
	     % subtracting one extra pulse if end phase larger than
	     % start phase, as this pulse will be takien care of in
	     % the "also added" 
	     variable subphas = (phtstop >= phtstart) ? 2 : 1;
	     if (int(nper)==1) subphas-- ;
	     ttot += tperbin*int(nper-subphas) ;
	     % ()=printf("added %.12fs to all bins for %d periods (subphase=%d)\n", tperbin, int(nper-subphas), subphas) ;
	     adddt+=tperbin*int(nper-subphas)*nbins ;
	   %take care of partial periods
	   %the beginning
	   ttot[[int(ceil(nstart)):nbins-1]] += tperbin ;
	     % ()=printf("also added %.12fs to bins %d:%d\n", tperbin, int(ceil(nstart)), nbins-1);
	     adddt+=tperbin*length([int(ceil(nstart)):nbins-1]) ;
	   %and at the end
	   ttot[[0:int(floor(nstop))-1 ]] += tperbin ;
	     % ()=printf("and also added %.12fs to bins %d:%d\n", tperbin, 0, int(floor(nstop))-1);
	     adddt+=tperbin*length([0:int(floor(nstop))-1 ]) ;
	  }

         %partial phase-bins can be taken care of in the same manner
	 %for all cases, so all the way at the end

	    %if nstart and nstop are in the same phase-bin, don't do that:
	 if (int(nstop)!=int(nstart) or nper>=1)
	      {
          % vmessage("taking now care of partial phase-bins") ;
		 if (nstart-int(nstart)>0)
		   {
		      ttot[[int(nstart)]] += (1-(nstart - int(nstart))) * tperbin ;
		      % ()=printf("added %.12f to bin %d\n", (1-(nstart - int(nstart))) * tperbin , int(nstart)) ;
		      adddt+= (1-(nstart - int(nstart))) * tperbin;
		   }
		 ttot[int(nstop)] += (nstop-int(nstop))*tperbin ;
		 % ()=printf("and added %.12f to bin %d\n", (nstop-int(nstop))*tperbin , int(nstop)) ;
		 adddt+=(nstop-int(nstop))*tperbin ;
	      }
	    
	    variable u ;
	    
	%filter out events outside of GTIs
	ndx = [where(tstart<=t<=tstop)] ;
	%some GTIs don't have any events, ignore them
	if(length(ndx)>0)
	   {
	      variable p_new = p+(pdot/2. + pddot*t[ndx]/6.)*t[ndx] ;
	      variable phas = ((t[ndx]) mod p_new)/p_new ;
	      if (length(where(phas<0))) phas[where(phas<0)] = 1+phas[where(phas<0)] ;
	      allphas[ndx] = phas ;
	      % print(phas) ;
	      % ()=printf("The actual periods we are using are between  P=%.7f--%.7f\n", p_new[0], p_new[-1]) ;
	      profile += histogram(phas,gphilo,gphihi, &rev_lo) ;
	      % if (length(phas)>30) _for u (0, 30, 1)  
	      % 	()=printf("ph=%.5f, t=%.5f, p=%.5f\n",phas[u], t[ndx[u]], p_new[u]) ;
	      % ()=printf("phase of first event %.4f\n", phas[0]) ;
	      % ()=printf("we now have %f photons in the profile (length(ndx)=%d)\n", sum(profile), length(ndx)) ;
	      
	   }
	    % vmessage("that should be all for GTI %d\n", g) ;
	    % vmessage("overall added %.7fs to ttot\n", adddt) ;
	 }
       
      variable x ;

      if (chatty>0){
	foreach x (wherenot(ttot))
	  {
	    ()=printf("Warning: phase bin %d contains no data\n", x);
	  }
      }

       if (qualifier_exists("returnphase"))
	 return struct{bin_lo=gphilo, bin_hi = gphihi , value=profile/ttot, error=sqrt(profile)/ttot, ttot = ttot, phas=allphas} ;
       else
	 return struct{bin_lo=gphilo, bin_hi = gphihi , value=profile/ttot, error=sqrt(profile)/ttot, ttot = ttot} ;
    }
   

  ifnot (qualifier_exists("exact")) {

      phi_lo = p + ( pdot/2. +  pddot*t/6.)*t ;  %calc phase

      lc = struct{t_lo=t, r=r, e=e,
	 per = phi_lo ,
	 phi_lo = ( t mod phi_lo ) / phi_lo ,  
	 phibin_lo = -ones(length(t))   } ;

      ndx = where( lc.phi_lo < 0. );
      if( length(ndx) > 0 )
   	{
      	lc.phi_lo[ndx]+= 1. ;
   	}

   
      h_lo= histogram(lc.phi_lo, gphilo, gphihi , &rev_lo) ;
      
      % ()=printf("doing it the usual way\n") ;   
      _for i (0, nbins-1, 1) {

	 ifnot (length(rev_lo[i])) {
	    if (chatty>0) {
	       ()=printf("Warning: phase bin %d contains no bins\n", i) ;
	    }
	    profile[i] = _NaN ;
	    npts[i] = 0 ;
	 }
	 
	 else {
	    ndx = rev_lo[i] ;
	    dte = dt[ndx] ;
	    ttot[i] = sum(dte) ;
	    npts[i] = h_lo[i] ;
	    profile[i] = sum((lc.r[ndx]*dte)) ;
	    lc.phibin_lo[ndx] = i ;
	    if (noerr) { proferr[i] = moment(lc.r[ndx]).var ; }
	    else {proferr[i] = sum((lc.e[ndx]*dte)^2) ; }
	 } 
      }
   
      % ()=printf("calculation loop took %fsec\n", toc ; ) ;
   
   }
   
   else {	   

      phi_lo = p + ( pdot/2. +  pddot*t/6.)*t ;  %calc phase
      phi_hi = p + ( pdot/2. +  pddot*(t+dt)/6.)*(t+dt) ;  %calc phase

      lc = struct{t_lo=t, t_hi = t+dt, r=r, e=e,
	 phi_lo = ( t mod phi_lo ) / phi_lo ,  
 	 phi_hi = ( (t+dt) mod phi_hi ) / phi_hi ,
	 phibin_lo = -ones(length(t)) ,
	 phibin_hi = -ones(length(t))   } ;

      ndx = where( lc.phi_lo < 0. );
      if( length(ndx) > 0 )
        {	
   	 lc.phi_lo[ndx]+= 1. ;
   	}
      ndx = where(  lc.phi_hi < 0. );
      if( length(ndx) > 0 )
        {	
   	 lc.phi_hi[ndx]+= 1. ;
        }
      
      h_lo= histogram(lc.phi_lo, gphilo, gphihi , &rev_lo) ;
      h_hi= histogram(lc.phi_hi, gphilo, gphihi , &rev_hi) ;
      
      % ()=printf("doing it the better way\n") ;   
  
      %save the corresponding phase bins for every time bin
      _for i (0, nbins-1, 1) {
	 rlo = rev_lo[i] ;
	 rhi = rev_hi[i] ;

	 lc.phibin_lo[rlo] = i ;
	 lc.phibin_hi[rhi] = i ;
      }
      
      %select the times where start and stop do not fall into the
      %same phase bin
      variable diffall = (lc.phibin_hi - lc.phibin_hi) ;
      % if (chatty>0) {()=printf("number of different entries: %d\n", length(diffall)) ; }

      variable bintime = double(p)/nbins ;

      variable phbin_lo, phbin_hi, diffphi_lo, diffphi_hi ;
  
      %()=printf("BINTIME = %.8f\n", bintime) ;
      
      _for i (0, length(lc.phibin_lo)-1, 1) 
	{
	   phbin_lo = lc.phibin_lo[i] ; %phasebin of start of time bin
	   phbin_hi = lc.phibin_hi[i] ; %phasebin of end of time bin
	   
	   if (phbin_lo == phbin_hi)
	     {
		npts[phbin_lo]+=1 ;
		ttot[phbin_lo]+=min([bintime,dt[i]]) ;
	     }
	   else 
	     {
	   variable phbin_hi2 = (phbin_hi == 0) ? nbins : phbin_hi ;

	   %()=printf("%d: looking at phase bins %d--%d\n", i, phbin_lo, phbin_hi2) ;
	   
	   if (phbin_lo <= phbin_hi2)
		  {
		
	   %maybe even more than a whole phasebin is covered by a time bin
	   npts[[phbin_lo+1:phbin_hi2-1]] += 1 ;
	   % print(npts) ;
	   ttot[[phbin_lo+1:phbin_hi2-1]] += bintime  ;
	   profile[[phbin_lo+1:phbin_hi2-1]] += lc.r[diffall[i]]*min([bintime,dt[i]]) ;
	   proferr[[phbin_lo+1:phbin_hi2-1]] += (lc.e[diffall[i]]*min([bintime,dt[i]]))^2 ;
		% ()=printf("added %.4f pts and %.14fs to phasbins %d--%d\n", 1, min([bintime,dt[i]]), phbin_lo+1,phbin_hi2-1) ;
		  }
		else 
		  {
		     npts[[[phbin_lo+1:length(npts)-1],[0:phbin_hi2-1]]] +=1 ;
		     ttot[[[phbin_lo+1:length(npts)-1],[0:phbin_hi2-1]]] += min([bintime,dt[i]]);
		  } 
		
		
      	   % %taking care of the first phasebin touched by the time bin
	   diffphi_lo = gphihi[phbin_lo] - lc.phi_lo[i] ; %difference in phi between border of phasebin and start of timebin
	   % ()=printf("gphihi[phbin_lo]= %.4f - lc.phi_lo[i]=%.5f =  diffph_lo = %.4f\n", gphihi[phbin_lo],lc.phi_lo[i],diffphi_lo) ;
	   npts[phbin_lo] += diffphi_lo*nbins ; %adding percentage of coverage to number of timebins in this phasebin
	   ttot[phbin_lo] += diffphi_lo * p ; %adding percentage of coverage of time
	   profile[phbin_lo] += lc.r[i]*diffphi_lo*p   ; %adding percentage of counts
	   proferr[phbin_lo] += (lc.e[i]*diffphi_lo*p)^2   ;
	   % ()=printf("added %.4f pts and %.4f s to phasbin %d\n", diffphi_lo*nbins, diffphi_lo * p, phbin_lo) ;
	
	   % %taking care of the last phasebin touched by the time bin
	   diffphi_hi = lc.phi_hi[i] - gphilo[phbin_hi] ; %difference in phi between border of phasebin and start of timebin
	   % ()=printf("diffph_hi = %.4f\n", diffphi_hi) ;
	   npts[phbin_hi] += diffphi_hi*nbins ; %adding percentage of coverage to number of timebins in this phasebin
	   ttot[phbin_hi] += diffphi_hi * p ; %adding percantage of coverage of time
	   profile[phbin_hi] += lc.r[i]*diffphi_hi*p  ;%adding percante of counts
	   proferr[phbin_hi] += (lc.e[i]*diffphi_hi*p)^2   ;
	   % ()=printf("added %.4f pts and %.4f s to phasbin %d\n", diffphi_hi*nbins, diffphi_hi * p, phbin_hi) ;
	     }
	   
	}
	   
	

      % _for i (0, nbins-1, 1) {

      % 	 ifnot (length(rev_lo[i])) {
      % 	    if (chatty>0) {
      % 	       ()=printf("Warning: phase bin %d contains no bins\n", i) ;
      % 	    }
      % 	    % profile[i] = _NaN ;
      % 	    % npts[i] = 0 ;
      % 	 }
	 
      % 	 else {
      % 	    ndx = rev_lo[i] ;
      % 	    dte = dt[ndx] ;
      % 	    % ttot[i] += sum(dte) ;
      % 	    npts[i] += h_lo[i] ;
      % 	    profile[i] += sum((lc.r[ndx]*dte)) ;
      % 	    if (noerr) {proferr[i] = moment(lc.r[ndx]).var ; }
      % 	    else {proferr[i] = sum((lc.e[ndx]*dte)^2) ; }
   }


   
	 % tic ;
	 % same =  where(lc.phibin_lo == lc.phibin_hi == i) ;
	 % % same = [0:100] ;
	 % ()=printf("sameing took %f sec\n", toc;) ; 
	 
	 % npts[i] = length(same) ;
	 % dte = lc.t_hi[same]-lc.t_lo[same] ;
	 % ttot[i] = sum(dte) ;
    	
	 % profile[i] = sum(lc.r[same]*dte) ;
	 % proferr[i] = sum((lc.e[same]*dte)^2) ;
      % }

  
      
  
   profile/=ttot ; %reform to countrate
   % proferr/=ttot ;
   ifnot (noerr) %{ proferr/=ttot;} 
    {  proferr = sqrt(proferr)/ttot ;  }
  
   if (qualifier_exists("returnlc")) 
     {return,struct{bin_lo=gphilo, bin_hi = gphihi , value=profile, error=proferr, npts=npts, ttot=ttot, lc=lc} ;}

   else{  return,struct{bin_lo=gphilo, bin_hi = gphihi , value=profile, error=proferr, npts=npts, ttot=ttot} ;}

}

