define hardnessratio()
%!%+
%\function{hardnessratio}
%\synopsis{calculates various X-ray hardness ratio from given count rates, including uncertainty if requested}
%\usage{Double_Type hardnessratio(Double_Type soft_count, Double_Type hard_count);}
%\qualifiers{
%  \qualifier{color}{: calculate the hardness ratio according to color=log10(S/H)}
%  \qualifier{hardness}{: calculate the hardness ratio according to hardness=(H-S)/(H+S)}
%  \qualifier{ratio}{: calculate the hardness ratio according to ratio=S/H (the default)}
%  \qualifier{back_s}{: background counts in the soft band}
%  \qualifier{back_h}{: background counts in the hard band}
%  \qualifier{backscale}{: ratio between the extraction regions for the source and the background}
%  \qualifier{exposure}{: Exposure per bin in seconds. If given, interpret soft_count, hard_count and t
%                         he backgrounds as rates. Multiply the rates with exposure to get the counts.}
%  \qualifier{backexposure}{: if given, the background exposure. Only taken into
%               account if exposure is also set. If not given, it is assumed that
%               the source and background exposures are identical.}
%  \qualifier{ratio_type}{: (DEPRECATED) Integer_Type, for hr=s/h choose 1,
%             for hr=(h-s)/(h+s) choose 2, for hr=log(s/h) choose 3, default = 1}
%  \qualifier{err_s}{: Array containing the uncertainties of the soft light curve}
%  \qualifier{err_h}{: Array containing the uncertainties of the hard light curve}
%}
%\description
%  This function calculates the so-called hardness ratio or color according to 
%  the three common prescriptions used in X-ray astronomy: S/H, (H-S)/(H+S), and log10(S/H).
%  If given, background counts are subtracted before the calculation, taking into account
%  different source and background extraction regions and different source and background exposure
%  times if necessary (if counts are given but the background exposure was different, set the
%  source exposure to 1 and the background exposure to the ratio of the source and background exposure
%  times. If rates are given, give the exposure times.
%
%  Error bars are calculated using Gaussian error propagation. Contrary to earlier versions of
%  this routine they are ALWAYS calculated!
%
%!%-
{
    variable s,h;

    switch(_NARGS)
    { case 2: (s,h) = (); }
    { help(_function_name()); return; }

    if (__is_numeric(s)!=2) {
	s=typecast(s,Double_Type);
    }
    if (__is_numeric(h)!=2) {
	h=typecast(h,Double_Type);
    }

    %
    % backwards compatibility
    %
    variable type=0;
    if (qualifier_exists("ratio_type")) {
	if (qualifier_exists("color") or qualifier_exists("ratio") or qualifier_exists("hardness")) {
	    throw UsageError,sprintf("%s: ratio_type cannot be used with the color, ratio, or hardness qualifiers.",_function_name());

	}
    }
    
    variable bs=qualifier("back_s",0.);
    variable bh=qualifier("back_h",0.);

    variable r=qualifier("backscale",1.);


    % Gehrels (1986) uncertainties
    variable sig_s = qualifier("err_s", sqrt(s+0.75)+1.);
    variable sig_h = qualifier("err_h", sqrt(h+0.75)+1.);

  
    %
    % exposure is given: convert rates to counts
    %
    if (qualifier_exists("exposure")) {
	variable expo=qualifier("exposure");
	s=s*expo;
	h=h*expo;

	variable backexpo=qualifier("backexposure",expo);
	bs=bs*backexpo;
	bh=bh*backexpo;

	r=r*expo/backexpo;

      % convert uncertainties from rates to counts
        if (qualifier_exists("err_s")){
	  sig_s = sig_s * expo;
        } else {
	  sig_s = sqrt(s+0.75)+1.;
          }

        if (qualifier_exists("err_h")){
          sig_h = sig_h * expo;
        } else {
	  sig_h = sqrt(h+0.75)+1.;
          }
     }
    
    % Gehrels (1986) uncertainties part 2
    variable sig_bs = (bs>0.) ? sqrt(bs+0.75)+1. : 0.;
    variable sig_bh = (bh>0.) ? sqrt(bh+0.75)+1. : 0.;

  
  
    variable dsbs=s-bs/r;
    variable dhbh=h-bh/r;

    variable sigs_sum=sqrt( sig_s^2.+(sig_bs/r)^2.);
    variable sigh_sum=sqrt( sig_h^2.+(sig_bh/r)^2.);

    % color: log10(S/H)
    if (qualifier_exists("color") || type==3) {
	variable color=log10(dsbs/dhbh);
	variable sig_color=sqrt( (sigs_sum/dsbs)^2. + (sigh_sum/dhbh)^2. )/log(10.);

	return (color,sig_color);
    }

    if (qualifier_exists("hardness") || type==2) {
	% hardness (H-S)/(H+S)
	variable hard=(h-s)/(h+s);
	variable sig_hard=2.*sqrt((dhbh*sigs_sum)^2.+(dsbs*sigh_sum)^2.)/(dhbh+dsbs)^2.;

	return (hard,sig_hard);
    }

    % default: S/H
    variable ratio=dsbs/dhbh;
    variable sig_ratio=ratio*sqrt( (sigs_sum/dsbs)^2. + (sigh_sum/dhbh)^2. );
    return (ratio,sig_ratio);

}



